import numpy as np


class TimeStepper:
    def __init__(self, exp, dexpinv, action):
        self.exp = exp
        self.dexpinv = dexpinv
        self.action = action
        self.a = None
        self.b = None
        self.c = None
        self.order = None
        self.s = None

    def step(self, f, t, y, h):
        n = y.size
        k_tilde = np.zeros((n, self.s))
        for i in range(self.s):
            u = np.zeros(n)
            for j in range(i - 1):
                # Explicit method, a is lower-triangular
                u += self.a[i, j] * k_tilde[:, j]
            k = f(t + self.c[i] * h, self.action(self.exp(h * u), y))
            k_tilde[:, i] = self.dexpinv(u, k, self.order)
        v = np.zeros(y.size)
        for i in range(self.s):
            v += self.b[i] * k_tilde[:, i]
        return self.action(self.exp(h * v), y)


class EulerLie(TimeStepper):
    def __init__(self, exp, dexpinv, action):
        super().__init__(exp, dexpinv, action)
        self.a = np.array([[0]])
        self.b = np.array([1])
        self.c = np.array([0])
        self.order = 1
        self.s = 1


class RKMK4(TimeStepper):
    def __init__(self, exp, dexpinv, action):
        super().__init__(exp, dexpinv, action)
        # fmt: off
        self.a = np.array(
            [
                [0,   0,   0,   0],
                [0.5, 0,   0,   0],
                [0,   0.5, 0,   0],
                [0,   0,   1.0, 0]
            ]
        )
        # fmt: on
        self.b = np.array([1 / 6, 1 / 3, 1 / 3, 1 / 6])
        self.c = np.array([0, 0.5, 0.5, 1.0])
        self.order = 4
        self.s = 4



from ..solve import solve
import numpy as np
import unittest


def A(t, y):
    return np.array(
        [[0, t, -0.4 * np.cos(t)], [-t, 0, 0.1 * t], [0.4 * np.cos(t), -0.1 * t, 0]]
    )


class Testso3(unittest.TestCase):
    def test_solve(self):
        y0 = [0.0, 0.0, 1.0]
        t_start = 0
        t_end = 5
        step_length = 0.01
        manifold = "hmnsphere"
        method = "RKMK4"
        solution = solve(A, y0, t_start, t_end, step_length, manifold, method)
        for i in range(len(solution.T)):
            self.assertAlmostEqual(np.linalg.norm(solution[:, i]), 1.0)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from .hmanifold import HomogenousManifold
from ..liealgebra import soLieAlgebra
from ..liegroup.liegroup import SOLieGroup


class HomogenousSphere(HomogenousManifold):
    def __init__(self, y):
        # Start with the unit vector pointing north
        if not isinstance(y, np.ndarray):
            try:
                y = np.array(y)
            except Exception:
                raise TypeError("y must be array_like")
        self.n = y.size
        self.y = y
        self.lie_group = SOLieGroup()
        self.lie_algebra = soLieAlgebra()
        super().__init__()

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        if value.size != self.n:
            raise ValueError("y does not have the correct dimension")
        elif not np.isclose(np.linalg.norm(value), 1.0):
            raise ValueError(
                f"y does not lie on the N-sphere. ||y|| should be one, was {np.linalg.norm(value)}" # noqa
            )
        self._y = value
